from __future__ import division

from ..tracker import GestureTracker, Gesture
from ..geometry import Positionable, MovingPositionable


class RotationGesture(Gesture, Positionable):
    """
    A rotation gesture has a angle in radians and a rotational centroid.
    """
    _type = 'rotate'

    def __init__(self, event, centroid, angle):
        Gesture.__init__(self, event)
        Positionable.__init__(self, *centroid.get_position())
        self.angle = angle

    def __str__(self):
        return '<RotationGesture at (%s, %s) angle=%s>' \
               % (self.x, self.y, self.angle)

    def get_angle(self):
        return self.angle


class PinchGesture(Gesture, Positionable):
    """
    A pinch gesture has a scale (1.0 means no scaling) and a centroid from
    which the scaling originates.
    """
    _type = 'pinch'

    def __init__(self, event, centroid, scale):
        Gesture.__init__(self, event)
        Positionable.__init__(self, *centroid.get_position())
        self.scale = scale

    def __str__(self):
        return '<PinchGesture at (%s, %s) scale=%s>' \
               % (self.x, self.y, self.scale)

    def get_scale(self):
        return self.scale


class DragGesture(Gesture, Positionable):
    """
    A momevent gesture has an initial position, and a translation from that
    position.
    """
    _type = 'drag'

    def __init__(self, event, initial_position, translation, n=1):
        Gesture.__init__(self, event)
        Positionable.__init__(self, *initial_position.get_position())
        self.translation = translation
        self.n = n

    def __str__(self):
        x, y = self
        tx, ty = self.translation
        return '<DragGesture at (%s, %s) translation=(%s, %s) n=%d>' \
               % (x, y, tx, ty, self.n)

    def get_translation(self):
        return self.translation


class TransformationTracker(GestureTracker):
    """
    Tracker for linear transformations. This implementation detects rotation,
    scaling and translation using the centroid of all touch points.
    """
    supported_gestures = [RotationGesture, PinchGesture, DragGesture]

    def __init__(self, widget):
        super(TransformationTracker, self).__init__(widget)

        # All touch points performing the transformation
        self.points = []

        # Current and previous centroid of all touch points
        self.prev_centroid = self.centroid = None

    def update_centroid(self):
        self.prev_centroid = self.centroid

        if not self.points:
            self.centroid = None
            return

        # Calculate average touch point coordinates
        l = len(self.points)
        coords = [p.get_position() for p in self.points]
        all_x, all_y = zip(*coords)
        x = sum(all_x) / l
        y = sum(all_y) / l

        # Update centroid positionable
        if self.centroid:
            self.centroid.set_position(x, y)
        else:
            self.centroid = MovingPositionable(x, y)

        #wx, wy = self.widget.get_screen_offset()
        #self.centroid.translate(-wx, -wy)

    def on_point_down(self, event):
        self.points.append(event.point)
        self.update_centroid()

    def on_point_move(self, event):
        point = event.point

        if point not in self.points:
            return

        l = len(self.points)

        if l > 1:
            # Rotation (around the previous centroid)
            rotation = point.rotation_around(self.centroid) / l
            self.trigger(RotationGesture(event, self.centroid, rotation))

            # Scale
            prev = point.get_previous_position().distance_to(self.centroid)
            dist = point.distance_to(self.centroid)
            dist = prev + (dist - prev) / l
            scale = dist / prev
            self.trigger(PinchGesture(event, self.centroid, scale))

            # Update centroid before movement can be detected
            self.update_centroid()

            # Movement using multiple touch points
            self.trigger(DragGesture(event, self.centroid,
                                     self.centroid.translation(), l))
        else:
            # Movement using one touch point
            self.trigger(DragGesture(event, point, point.translation()))

    def on_point_up(self, event):
        if event.point not in self.points:
            return

        self.points.remove(event.point)
        self.update_centroid()
